"""Decorators for ``cmd2`` commands."""

import argparse
import functools
from collections.abc import (
    Callable,
    Sequence,
)
from typing import (
    TYPE_CHECKING,
    Any,
    TypeAlias,
    TypeVar,
    overload,
)

from . import constants
from .argparse_utils import (
    ArgparseCommandSpec,
    ClassParamParserFactory,
    Cmd2ArgumentParser,
    NoParamParserFactory,
    ParserSource,
    SubcommandSpec,
)
from .command_set import CommandSet
from .exceptions import Cmd2ArgparseError
from .parsing import Statement
from .types import (
    CmdOrSetT,
    UnboundCommandFunc,
)

if TYPE_CHECKING:  # pragma: no cover
    from .cmd2 import Cmd

F = TypeVar("F", bound=Callable[..., Any])


def with_category(
    category: str,
) -> Callable[[F], F]:
    """Decorate a ``do_*`` command function to apply a category.

    Permissive type hints allow this decorator to be stacked in any order, even
    when other decorators in the chain transform the signature or return type of
    the command function.

    :param category: the name of the category in which this command should
                     be grouped when displaying the list of commands.
    :return: a decorator that assigns the specified category to the command function

    Example:
    ```py
    class MyApp(cmd2.Cmd):
        @cmd2.with_category("Text Functions")
        def do_echo(self, args: cmd2.Statement) -> None:
            self.poutput(args)
    ```

    For an alternative approach to categorizing commands using a function, see
    [cmd2.utils.categorize][]

    """

    def cat_decorator(func: F) -> F:
        from .utils import categorize

        categorize(func, category)
        return func

    return cat_decorator


##########################
# The _parse_positionals and _arg_swap functions allow for additional positional args to be preserved
# in cmd2 command functions/callables. As long as the 2-ple of arguments we expect to be there can be
# found we can swap out the statement with each decorator's specific parameters
##########################
def _parse_positionals(args: tuple[Any, ...]) -> tuple["Cmd", Statement | str]:
    """Inspect the positional arguments until the cmd2.Cmd argument is found.

    Assumes that we will find cmd2.Cmd followed by the command statement object or string.

    :arg args: The positional arguments to inspect
    :return: The cmd2.Cmd reference and the command line statement.
    """
    for pos, arg in enumerate(args):
        from .cmd2 import Cmd

        if isinstance(arg, (Cmd, CommandSet)) and len(args) > pos + 1:
            if isinstance(arg, CommandSet):
                arg = arg._cmd  # noqa: PLW2901
            next_arg = args[pos + 1]
            if isinstance(next_arg, (Statement, str)):
                return arg, args[pos + 1]

    # This shouldn't happen unless we forget to pass statement in `Cmd.onecmd` or
    # somehow call the unbound class method.
    raise TypeError("Expected arguments: cmd: cmd2.Cmd, statement: Union[Statement, str] Not found")


def _arg_swap(args: Sequence[Any], search_arg: Any, *replace_arg: Any) -> list[Any]:
    """Swap the Statement parameter with one or more decorator-specific parameters.

    :param args: The original positional arguments
    :param search_arg: The argument to search for (usually the Statement)
    :param replace_arg: The arguments to substitute in
    :return: The new set of arguments to pass to the command function
    """
    index = args.index(search_arg)
    args_list = list(args)
    args_list[index : index + 1] = replace_arg
    return args_list


# The standard cmd2 command function signature (e.g. do_command(self, statement))
RawCommandFunc: TypeAlias = UnboundCommandFunc[CmdOrSetT, [Statement | str]]


# Function signature for a command function that accepts a pre-processed argument list from user input
ArgListCommandFunc: TypeAlias = UnboundCommandFunc[CmdOrSetT, [list[str]]]


# Overload for: @with_argument_list
@overload
def with_argument_list(
    cmd_func: ArgListCommandFunc[CmdOrSetT],
    *,
    preserve_quotes: bool = False,
) -> RawCommandFunc[CmdOrSetT]: ...


# Overload for: @with_argument_list(preserve_quotes=True)
@overload
def with_argument_list(
    cmd_func: None = None,
    *,
    preserve_quotes: bool = False,
) -> Callable[[ArgListCommandFunc[CmdOrSetT]], RawCommandFunc[CmdOrSetT]]: ...


def with_argument_list(
    cmd_func: ArgListCommandFunc[CmdOrSetT] | None = None,
    *,
    preserve_quotes: bool = False,
) -> RawCommandFunc[CmdOrSetT] | Callable[[ArgListCommandFunc[CmdOrSetT]], RawCommandFunc[CmdOrSetT]]:
    """Decorate a ``do_*`` command function to receive a list of parsed arguments.

    This decorator can be used either directly (``@with_argument_list``) or as a
    factory with arguments (``@with_argument_list(preserve_quotes=True)``).

    :param cmd_func: The command function being decorated.
    :param preserve_quotes: If ``True``, argument quotes will not be stripped from the input.
    :return: A command function that accepts a list of strings instead of a raw string.

    Example:
    ```py
    class MyApp(cmd2.Cmd):
        # Basic usage: receives a list of words with quotes stripped
        @cmd2.with_argument_list
        def do_echo(self, arglist: list[str]) -> None:
            self.poutput(" ".join(arglist))

        # Factory usage: preserves quotes in the argument list
        @cmd2.with_argument_list(preserve_quotes=True)
        def do_print_raw(self, arglist: list[str]) -> None:
            self.poutput(" ".join(arglist))
    ```

    """

    def arg_decorator(func: ArgListCommandFunc[CmdOrSetT]) -> RawCommandFunc[CmdOrSetT]:
        """Decorate function that ingests an Argument List function and returns a raw command function.

        The returned function will process the raw input into an argument list to be passed to the wrapped function.

        :param func: The defined argument list command function
        :return: Function that takes raw input and converts to an argument list to pass to the wrapped function.
        """

        @functools.wraps(func)
        def cmd_wrapper(*args: Any, **kwargs: Any) -> bool | None:
            """Command function wrapper which translates command line into an argument list and calls actual command function.

            :param args: All positional arguments to this function.  We're expecting there to be:
                            cmd_app, statement: Union[Statement, str]
                            contiguously somewhere in the list
            :param kwargs: any keyword arguments being passed to command function
            :return: return value of command function
            """
            cmd_app, statement = _parse_positionals(args)
            _, command_arg_list = cmd_app.statement_parser.get_command_arg_list(command_name, statement, preserve_quotes)
            func_arg_list = _arg_swap(args, statement, command_arg_list)
            return func(*func_arg_list, **kwargs)

        command_name = func.__name__[len(constants.COMMAND_FUNC_PREFIX) :]
        cmd_wrapper.__doc__ = func.__doc__
        return cmd_wrapper

    if callable(cmd_func):
        return arg_decorator(cmd_func)
    return arg_decorator


# Function signatures for command functions that use a Cmd2ArgumentParser to process user input
ArgparseCommandFunc: TypeAlias = (
    # (self, args: argparse.Namespace)
    UnboundCommandFunc[CmdOrSetT, [argparse.Namespace]]
    # (self, args: argparse.Namespace, unknown_args: list[str])
    | UnboundCommandFunc[CmdOrSetT, [argparse.Namespace, list[str]]]
)


# Overload for: existing parser instance
@overload
def with_argparser(
    parser_source: Cmd2ArgumentParser,
    *,
    ns_provider: Callable[..., argparse.Namespace] | None = None,
    preserve_quotes: bool = False,
    with_unknown_args: bool = False,
) -> Callable[[ArgparseCommandFunc[CmdOrSetT]], RawCommandFunc[CmdOrSetT]]: ...


# Overload for: factory with no arguments (including staticmethod)
@overload
def with_argparser(
    parser_source: NoParamParserFactory,
    *,
    ns_provider: Callable[..., argparse.Namespace] | None = None,
    preserve_quotes: bool = False,
    with_unknown_args: bool = False,
) -> Callable[[ArgparseCommandFunc[CmdOrSetT]], RawCommandFunc[CmdOrSetT]]: ...


# Overload for: factory with a class argument (including classmethod)
@overload
def with_argparser(
    parser_source: ClassParamParserFactory[CmdOrSetT],
    *,
    ns_provider: Callable[..., argparse.Namespace] | None = None,
    preserve_quotes: bool = False,
    with_unknown_args: bool = False,
) -> Callable[[ArgparseCommandFunc[CmdOrSetT]], RawCommandFunc[CmdOrSetT]]: ...


def with_argparser(
    parser_source: ParserSource[CmdOrSetT],
    *,
    ns_provider: Callable[..., argparse.Namespace] | None = None,
    preserve_quotes: bool = False,
    with_unknown_args: bool = False,
) -> Callable[[ArgparseCommandFunc[CmdOrSetT]], RawCommandFunc[CmdOrSetT]]:
    """Decorate a ``do_*`` command function to populate its ``args`` argument with a Cmd2ArgumentParser.

    :param parser_source: an existing Cmd2ArgumentParser instance or a factory
                          (callable, staticmethod, or classmethod) that returns one.
    :param ns_provider: An optional function that accepts a cmd2.Cmd or cmd2.CommandSet object as an argument and returns an
                        argparse.Namespace. This is useful if the Namespace needs to be prepopulated with state data that
                        affects parsing.
    :param preserve_quotes: if ``True``, then arguments passed to argparse maintain their quotes
    :param with_unknown_args: if true, then capture unknown args
    :return: function that gets passed argparse-parsed args in a ``Namespace``
             A ``cmd2_statement`` attribute is included in the ``Namespace`` to provide access to the
             [cmd2.Statement][] object that was created when parsing the command line. This can be useful
             if the command function needs to know the command line.

    Example:
    ```py
    parser = cmd2.Cmd2ArgumentParser()
    parser.add_argument("-p", "--piglatin", action="store_true", help="atinLay")
    parser.add_argument("-s", "--shout", action="store_true", help="N00B EMULATION MODE")
    parser.add_argument("-r", "--repeat", type=int, help="output [n] times")
    parser.add_argument("words", nargs="+", help="words to print")


    class MyApp(cmd2.Cmd):
        @cmd2.with_argparser(parser, preserve_quotes=True)
        def do_argprint(self, args: argparse.Namespace) -> None:
            "Print the options and argument list this options command was called with."
            self.poutput(f"args: {args!r}")
    ```

    Example with unknown args:

    ```py
    parser = cmd2.Cmd2ArgumentParser()
    parser.add_argument("-p", "--piglatin", action="store_true", help="atinLay")
    parser.add_argument("-s", "--shout", action="store_true", help="N00B EMULATION MODE")
    parser.add_argument("-r", "--repeat", type=int, help="output [n] times")


    class MyApp(cmd2.Cmd):
        @cmd2.with_argparser(parser, with_unknown_args=True)
        def do_argprint(self, args: argparse.Namespace, unknown_args: list[str]):
            "Print the options and argument list this options command was called with."
            self.poutput(f"args: {args!r}")
            self.poutput(f"unknown_args: {unknown_args}")
    ```

    """

    def arg_decorator(func: ArgparseCommandFunc[CmdOrSetT]) -> RawCommandFunc[CmdOrSetT]:
        """Decorate function that ingests an Argparse Command Function and returns a raw command function.

        The returned function will process the raw input into an argparse Namespace to be passed to the wrapped function.

        :param func: The defined argparse command function
        :return: Function that takes raw input and converts to an argparse Namespace to passed to the wrapped function.
        """

        @functools.wraps(func)  # type: ignore[arg-type, ty:invalid-argument-type]
        def cmd_wrapper(*args: Any, **kwargs: Any) -> bool | None:
            """Command function wrapper which translates command line into argparse Namespace and call actual command function.

            :param args: All positional arguments to this function.  We're expecting there to be:
                            cmd_app, statement: Union[Statement, str]
                            contiguously somewhere in the list
            :param kwargs: any keyword arguments being passed to command function
            :return: return value of command function
            :raises Cmd2ArgparseError: if argparse has error parsing command line
            """
            cmd_app, statement_arg = _parse_positionals(args)
            statement, command_arg_list = cmd_app.statement_parser.get_command_arg_list(
                command_name, statement_arg, preserve_quotes
            )

            # Pass cmd_wrapper instead of func, since it contains the parser info.
            arg_parser = cmd_app.command_parsers.get(cmd_wrapper)
            if arg_parser is None:
                # This shouldn't be possible to reach
                raise ValueError(f"No argument parser found for {command_name}")  # pragma: no cover

            if ns_provider is None:
                initial_namespace = None
            else:
                # The namespace provider may or may not be defined in the same class as the command. Since provider
                # functions are registered with the command argparser before anything is instantiated, we
                # need to find an instance at runtime that matches the types during declaration
                provider_self = cmd_app._resolve_func_self(ns_provider, args[0])
                initial_namespace = ns_provider(provider_self if provider_self is not None else cmd_app)

            try:
                parsing_results: tuple[argparse.Namespace] | tuple[argparse.Namespace, list[str]]
                with arg_parser.output_to(cmd_app.stdout):
                    if with_unknown_args:
                        parsing_results = arg_parser.parse_known_args(command_arg_list, initial_namespace)  # type: ignore[assignment, ty:invalid-assignment]
                    else:
                        parsing_results = (arg_parser.parse_args(command_arg_list, initial_namespace),)  # type: ignore[assignment, ty:invalid-assignment]
            except SystemExit as exc:
                raise Cmd2ArgparseError from exc

            # Add cmd2-specific attributes to the Namespace
            parsed_namespace = parsing_results[0]

            # Include the Statement object created from the command line
            setattr(parsed_namespace, constants.NS_ATTR_STATEMENT, statement)

            # Ensure subcommand function attribute is always present.
            if not hasattr(parsed_namespace, constants.NS_ATTR_SUBCOMMAND_FUNC):
                setattr(parsed_namespace, constants.NS_ATTR_SUBCOMMAND_FUNC, None)

            func_arg_list = _arg_swap(args, statement_arg, *parsing_results)
            return func(*func_arg_list, **kwargs)

        command_name = func.__name__[len(constants.COMMAND_FUNC_PREFIX) :]

        spec = ArgparseCommandSpec(
            parser_source=parser_source,
            preserve_quotes=preserve_quotes,
        )
        setattr(cmd_wrapper, constants.ARGPARSE_COMMAND_ATTR_SPEC, spec)

        return cmd_wrapper

    return arg_decorator


# Overload for: existing parser instance
@overload
def as_subcommand_to(
    command: str,
    subcommand: str,
    parser_source: Cmd2ArgumentParser,
    *,
    help: str | None = None,
    aliases: Sequence[str] = (),
    deprecated: bool = False,
) -> Callable[[F], F]: ...


# Overload for: factory with no arguments (including staticmethod)
@overload
def as_subcommand_to(
    command: str,
    subcommand: str,
    parser_source: NoParamParserFactory,
    *,
    help: str | None = None,
    aliases: Sequence[str] = (),
    deprecated: bool = False,
) -> Callable[[F], F]: ...


# Overload for: factory with a class argument (including classmethod)
@overload
def as_subcommand_to(
    command: str,
    subcommand: str,
    parser_source: ClassParamParserFactory[CmdOrSetT],
    *,
    help: str | None = None,
    aliases: Sequence[str] = (),
    deprecated: bool = False,
) -> Callable[[F], F]: ...


def as_subcommand_to(
    command: str,
    subcommand: str,
    parser_source: ParserSource[CmdOrSetT],
    *,
    help: str | None = None,  # noqa: A002
    aliases: Sequence[str] = (),
    deprecated: bool = False,
) -> Callable[[F], F]:
    """Tag a function as a subcommand to an existing argparse decorated command.

    Permissive type hints allow this decorator to be stacked in any order, even
    when other decorators in the chain transform the signature or return type of
    the subcommand function.

    While this decorator has permissive type hints, the subcommand function's signature
    must match the root command's signature. For example, if the root command uses
    `with_unknown_args=True`, then the subcommand function must also accept the
    unknown arguments list.

    :param command: Command Name. Space-delimited subcommands may optionally be specified
    :param subcommand: Subcommand name
    :param parser_source: an existing Cmd2ArgumentParser instance or a factory
                          (callable, staticmethod, or classmethod) that returns one.
    :param help: optional help message for this subcommand which displays in the list of subcommands
                 of the command we are adding to.
    :param aliases: optional alternative names for this subcommand.
    :param deprecated: whether this subcommand is deprecated (requires Python 3.13+).
    :return: a decorator which configures the target function to be a subcommand handler

    Example:
    ```py
    base_parser = cmd2.Cmd2ArgumentParser()
    base_parser.add_subparsers(title="subcommands", metavar="SUBCOMMAND", required=True)
    sub_parser = cmd2.Cmd2ArgumentParser()


    class MyApp(cmd2.Cmd):
        @cmd2.with_argparser(base_parser)
        def do_base(self, args: argparse.Namespace) -> None:
            args.cmd2_subcommand_func(args)

        @cmd2.as_subcommand_to("base", "sub", sub_parser, help="the subcommand")
        def sub_func(self, args: argparse.Namespace) -> None:
            self.poutput("Subcommand executed")
    ```

    """

    def arg_decorator(func: F) -> F:
        spec = SubcommandSpec(
            name=subcommand,
            command=command,
            help=help,
            aliases=tuple(aliases),
            deprecated=deprecated,
            parser_source=parser_source,
        )
        setattr(func, constants.SUBCOMMAND_ATTR_SPEC, spec)
        return func

    return arg_decorator



"""Import certain things for backwards compatibility."""

import contextlib
import importlib.metadata as importlib_metadata

with contextlib.suppress(importlib_metadata.PackageNotFoundError):
    __version__ = importlib_metadata.version(__name__)

from . import (
    plugin,
    rich_utils,
    string_utils,
)
from .annotated import (
    ArgumentBlock,
    with_annotated,
)
from .argparse_completer import set_default_argparse_completer
from .argparse_utils import (
    Cmd2ArgumentParser,
    SubcommandRecord,
    register_argparse_argument_parameter,
    set_default_argument_parser,
)
from .cmd2 import Cmd
from .colors import Color
from .command_set import CommandSet
from .completion import (
    Choices,
    CompletionItem,
    Completions,
)
from .constants import (
    COMMAND_NAME,
    DEFAULT_SHORTCUTS,
)
from .decorators import (
    as_subcommand_to,
    with_argparser,
    with_argument_list,
    with_category,
)
from .exceptions import (
    Cmd2ArgparseError,
    CommandSetRegistrationError,
    CompletionError,
    PassThroughException,
    SkipPostcommandHooks,
)
from .parsing import Statement
from .py_bridge import CommandResult
from .rich_utils import (
    ArgumentDefaultsCmd2HelpFormatter,
    Cmd2HelpFormatter,
    MetavarTypeCmd2HelpFormatter,
    RawDescriptionCmd2HelpFormatter,
    RawTextCmd2HelpFormatter,
    TextGroup,
)
from .string_utils import stylize
from .styles import Cmd2Style
from .theme import (
    get_theme,
    reset_theme,
    update_theme,
)
from .utils import (
    CustomCompletionSettings,
    Settable,
    categorize,
    set_default_str_sort_key,
)

__all__: list[str] = [  # noqa: RUF022
    "COMMAND_NAME",
    "DEFAULT_SHORTCUTS",
    # Argparse Exports
    "Cmd2ArgumentParser",
    "SubcommandRecord",
    "register_argparse_argument_parameter",
    "set_default_argparse_completer",
    "set_default_argument_parser",
    # Cmd2
    "Cmd",
    "CommandResult",
    "CommandSet",
    "Statement",
    # Colors
    "Color",
    # Completion
    "Choices",
    "CompletionItem",
    "Completions",
    # Annotated commands
    "ArgumentBlock",
    # Decorators
    "with_annotated",
    "with_argument_list",
    "with_argparser",
    "with_category",
    "as_subcommand_to",
    # Exceptions
    "Cmd2ArgparseError",
    "CommandSetRegistrationError",
    "CompletionError",
    "PassThroughException",
    "SkipPostcommandHooks",
    # modules
    "plugin",
    "rich_utils",
    "string_utils",
    # Rich Utils
    "ArgumentDefaultsCmd2HelpFormatter",
    "Cmd2HelpFormatter",
    "MetavarTypeCmd2HelpFormatter",
    "RawDescriptionCmd2HelpFormatter",
    "RawTextCmd2HelpFormatter",
    "TextGroup",
    # String Utils
    "stylize",
    # Styles
    "Cmd2Style",
    # Theme
    "get_theme",
    "reset_theme",
    "update_theme",
    # Utilities
    "categorize",
    "CustomCompletionSettings",
    "Settable",
    "set_default_str_sort_key",
]



"""Provides a convenient StrEnum for Rich color names."""

from enum import StrEnum


class Color(StrEnum):
    """An enumeration of all color names supported by the Rich library.

    Using this enum allows for autocompletion and prevents typos when referencing
    color names. The members can be used for both foreground and background colors.

    Aside from DEFAULT, these colors come from the rich.color.ANSI_COLOR_NAMES dictionary.

    Note: The terminal color settings determines the appearance of the follow 16 colors.

    |                |               |
    |----------------|---------------|
    | BLACK          | BRIGHT_WHITE  |
    | BLUE           | BRIGHT_YELLOW |
    | BRIGHT_BLACK   | CYAN          |
    | BRIGHT_BLUE    | GREEN         |
    | BRIGHT_CYAN    | MAGENTA       |
    | BRIGHT_GREEN   | RED           |
    | BRIGHT_MAGENTA | WHITE         |
    | BRIGHT_RED     | YELLOW        |
    """

    DEFAULT = "default"
    """Represents the terminal's default foreground or background color."""

    AQUAMARINE1 = "aquamarine1"
    AQUAMARINE3 = "aquamarine3"
    BLACK = "black"
    BLUE = "blue"
    BLUE1 = "blue1"
    BLUE3 = "blue3"
    BLUE_VIOLET = "blue_violet"
    BRIGHT_BLACK = "bright_black"
    BRIGHT_BLUE = "bright_blue"
    BRIGHT_CYAN = "bright_cyan"
    BRIGHT_GREEN = "bright_green"
    BRIGHT_MAGENTA = "bright_magenta"
    BRIGHT_RED = "bright_red"
    BRIGHT_WHITE = "bright_white"
    BRIGHT_YELLOW = "bright_yellow"
    CADET_BLUE = "cadet_blue"
    CHARTREUSE1 = "chartreuse1"
    CHARTREUSE2 = "chartreuse2"
    CHARTREUSE3 = "chartreuse3"
    CHARTREUSE4 = "chartreuse4"
    CORNFLOWER_BLUE = "cornflower_blue"
    CORNSILK1 = "cornsilk1"
    CYAN = "cyan"
    CYAN1 = "cyan1"
    CYAN2 = "cyan2"
    CYAN3 = "cyan3"
    DARK_BLUE = "dark_blue"
    DARK_CYAN = "dark_cyan"
    DARK_GOLDENROD = "dark_goldenrod"
    DARK_GREEN = "dark_green"
    DARK_KHAKI = "dark_khaki"
    DARK_MAGENTA = "dark_magenta"
    DARK_OLIVE_GREEN1 = "dark_olive_green1"
    DARK_OLIVE_GREEN2 = "dark_olive_green2"
    DARK_OLIVE_GREEN3 = "dark_olive_green3"
    DARK_ORANGE = "dark_orange"
    DARK_ORANGE3 = "dark_orange3"
    DARK_RED = "dark_red"
    DARK_SEA_GREEN = "dark_sea_green"
    DARK_SEA_GREEN1 = "dark_sea_green1"
    DARK_SEA_GREEN2 = "dark_sea_green2"
    DARK_SEA_GREEN3 = "dark_sea_green3"
    DARK_SEA_GREEN4 = "dark_sea_green4"
    DARK_SLATE_GRAY1 = "dark_slate_gray1"
    DARK_SLATE_GRAY2 = "dark_slate_gray2"
    DARK_SLATE_GRAY3 = "dark_slate_gray3"
    DARK_TURQUOISE = "dark_turquoise"
    DARK_VIOLET = "dark_violet"
    DEEP_PINK1 = "deep_pink1"
    DEEP_PINK2 = "deep_pink2"
    DEEP_PINK3 = "deep_pink3"
    DEEP_PINK4 = "deep_pink4"
    DEEP_SKY_BLUE1 = "deep_sky_blue1"
    DEEP_SKY_BLUE2 = "deep_sky_blue2"
    DEEP_SKY_BLUE3 = "deep_sky_blue3"
    DEEP_SKY_BLUE4 = "deep_sky_blue4"
    DODGER_BLUE1 = "dodger_blue1"
    DODGER_BLUE2 = "dodger_blue2"
    DODGER_BLUE3 = "dodger_blue3"
    GOLD1 = "gold1"
    GOLD3 = "gold3"
    GRAY0 = "gray0"
    GRAY3 = "gray3"
    GRAY7 = "gray7"
    GRAY11 = "gray11"
    GRAY15 = "gray15"
    GRAY19 = "gray19"
    GRAY23 = "gray23"
    GRAY27 = "gray27"
    GRAY30 = "gray30"
    GRAY35 = "gray35"
    GRAY37 = "gray37"
    GRAY39 = "gray39"
    GRAY42 = "gray42"
    GRAY46 = "gray46"
    GRAY50 = "gray50"
    GRAY53 = "gray53"
    GRAY54 = "gray54"
    GRAY58 = "gray58"
    GRAY62 = "gray62"
    GRAY63 = "gray63"
    GRAY66 = "gray66"
    GRAY69 = "gray69"
    GRAY70 = "gray70"
    GRAY74 = "gray74"
    GRAY78 = "gray78"
    GRAY82 = "gray82"
    GRAY84 = "gray84"
    GRAY85 = "gray85"
    GRAY89 = "gray89"
    GRAY93 = "gray93"
    GRAY100 = "gray100"
    GREEN = "green"
    GREEN1 = "green1"
    GREEN3 = "green3"
    GREEN4 = "green4"
    GREEN_YELLOW = "green_yellow"
    GREY0 = "grey0"
    GREY3 = "grey3"
    GREY7 = "grey7"
    GREY11 = "grey11"
    GREY15 = "grey15"
    GREY19 = "grey19"
    GREY23 = "grey23"
    GREY27 = "grey27"
    GREY30 = "grey30"
    GREY35 = "grey35"
    GREY37 = "grey37"
    GREY39 = "grey39"
    GREY42 = "grey42"
    GREY46 = "grey46"
    GREY50 = "grey50"
    GREY53 = "grey53"
    GREY54 = "grey54"
    GREY58 = "grey58"
    GREY62 = "grey62"
    GREY63 = "grey63"
    GREY66 = "grey66"
    GREY69 = "grey69"
    GREY70 = "grey70"
    GREY74 = "grey74"
    GREY78 = "grey78"
    GREY82 = "grey82"
    GREY84 = "grey84"
    GREY85 = "grey85"
    GREY89 = "grey89"
    GREY93 = "grey93"
    GREY100 = "grey100"
    HONEYDEW2 = "honeydew2"
    HOT_PINK = "hot_pink"
    HOT_PINK2 = "hot_pink2"
    HOT_PINK3 = "hot_pink3"
    INDIAN_RED = "indian_red"
    INDIAN_RED1 = "indian_red1"
    KHAKI1 = "khaki1"
    KHAKI3 = "khaki3"
    LIGHT_CORAL = "light_coral"
    LIGHT_CYAN1 = "light_cyan1"
    LIGHT_CYAN3 = "light_cyan3"
    LIGHT_GOLDENROD1 = "light_goldenrod1"
    LIGHT_GOLDENROD2 = "light_goldenrod2"
    LIGHT_GOLDENROD3 = "light_goldenrod3"
    LIGHT_GREEN = "light_green"
    LIGHT_PINK1 = "light_pink1"
    LIGHT_PINK3 = "light_pink3"
    LIGHT_PINK4 = "light_pink4"
    LIGHT_SALMON1 = "light_salmon1"
    LIGHT_SALMON3 = "light_salmon3"
    LIGHT_SEA_GREEN = "light_sea_green"
    LIGHT_SKY_BLUE1 = "light_sky_blue1"
    LIGHT_SKY_BLUE3 = "light_sky_blue3"
    LIGHT_SLATE_BLUE = "light_slate_blue"
    LIGHT_SLATE_GRAY = "light_slate_gray"
    LIGHT_SLATE_GREY = "light_slate_grey"
    LIGHT_STEEL_BLUE = "light_steel_blue"
    LIGHT_STEEL_BLUE1 = "light_steel_blue1"
    LIGHT_STEEL_BLUE3 = "light_steel_blue3"
    LIGHT_YELLOW3 = "light_yellow3"
    MAGENTA = "magenta"
    MAGENTA1 = "magenta1"
    MAGENTA2 = "magenta2"
    MAGENTA3 = "magenta3"
    MEDIUM_ORCHID = "medium_orchid"
    MEDIUM_ORCHID1 = "medium_orchid1"
    MEDIUM_ORCHID3 = "medium_orchid3"
    MEDIUM_PURPLE = "medium_purple"
    MEDIUM_PURPLE1 = "medium_purple1"
    MEDIUM_PURPLE2 = "medium_purple2"
    MEDIUM_PURPLE3 = "medium_purple3"
    MEDIUM_PURPLE4 = "medium_purple4"
    MEDIUM_SPRING_GREEN = "medium_spring_green"
    MEDIUM_TURQUOISE = "medium_turquoise"
    MEDIUM_VIOLET_RED = "medium_violet_red"
    MISTY_ROSE1 = "misty_rose1"
    MISTY_ROSE3 = "misty_rose3"
    NAVAJO_WHITE1 = "navajo_white1"
    NAVAJO_WHITE3 = "navajo_white3"
    NAVY_BLUE = "navy_blue"
    ORANGE1 = "orange1"
    ORANGE3 = "orange3"
    ORANGE4 = "orange4"
    ORANGE_RED1 = "orange_red1"
    ORCHID = "orchid"
    ORCHID1 = "orchid1"
    ORCHID2 = "orchid2"
    PALE_GREEN1 = "pale_green1"
    PALE_GREEN3 = "pale_green3"
    PALE_TURQUOISE1 = "pale_turquoise1"
    PALE_TURQUOISE4 = "pale_turquoise4"
    PALE_VIOLET_RED1 = "pale_violet_red1"
    PINK1 = "pink1"
    PINK3 = "pink3"
    PLUM1 = "plum1"
    PLUM2 = "plum2"
    PLUM3 = "plum3"
    PLUM4 = "plum4"
    PURPLE = "purple"
    PURPLE3 = "purple3"
    PURPLE4 = "purple4"
    RED = "red"
    RED1 = "red1"
    RED3 = "red3"
    ROSY_BROWN = "rosy_brown"
    ROYAL_BLUE1 = "royal_blue1"
    SALMON1 = "salmon1"
    SANDY_BROWN = "sandy_brown"
    SEA_GREEN1 = "sea_green1"
    SEA_GREEN2 = "sea_green2"
    SEA_GREEN3 = "sea_green3"
    SKY_BLUE1 = "sky_blue1"
    SKY_BLUE2 = "sky_blue2"
    SKY_BLUE3 = "sky_blue3"
    SLATE_BLUE1 = "slate_blue1"
    SLATE_BLUE3 = "slate_blue3"
    SPRING_GREEN1 = "spring_green1"
    SPRING_GREEN2 = "spring_green2"
    SPRING_GREEN3 = "spring_green3"
    SPRING_GREEN4 = "spring_green4"
    STEEL_BLUE = "steel_blue"
    STEEL_BLUE1 = "steel_blue1"
    STEEL_BLUE3 = "steel_blue3"
    TAN = "tan"
    THISTLE1 = "thistle1"
    THISTLE3 = "thistle3"
    TURQUOISE2 = "turquoise2"
    TURQUOISE4 = "turquoise4"
    VIOLET = "violet"
    WHEAT1 = "wheat1"
    WHEAT4 = "wheat4"
    WHITE = "white"
    YELLOW = "yellow"
    YELLOW1 = "yellow1"
    YELLOW2 = "yellow2"
    YELLOW3 = "yellow3"
    YELLOW4 = "yellow4"



"""Provides string utility functions.

This module offers a collection of string utility functions built on the Rich library.
These utilities are designed to correctly handle strings with ANSI style sequences and
full-width characters (like those used in CJK languages).
"""

from collections.abc import Sequence

from rich.align import AlignMethod
from rich.style import StyleType
from rich.text import Text

from . import rich_utils as ru


def align(
    val: str,
    align: AlignMethod,
    width: int | None = None,
    character: str = " ",
) -> str:
    """Align string to a given width.

    There are convenience wrappers around this function: align_left(), align_center(), and align_right()

    :param val: string to align
    :param align: one of "left", "center", or "right".
    :param width: Desired width. Defaults to width of the terminal.
    :param character: Character to pad with. Defaults to " ".

    """
    if width is None:
        width = ru.console_width()

    text = Text.from_ansi(val)
    text.align(align, width=width, character=character)
    return ru.rich_text_to_string(text)


def align_left(
    val: str,
    width: int | None = None,
    character: str = " ",
) -> str:
    """Left-align string to a given width.

    :param val: string to align
    :param width: Desired width. Defaults to width of the terminal.
    :param character: Character to pad with. Defaults to " ".

    """
    return align(val, "left", width=width, character=character)


def align_center(
    val: str,
    width: int | None = None,
    character: str = " ",
) -> str:
    """Center-align string to a given width.

    :param val: string to align
    :param width: Desired width. Defaults to width of the terminal.
    :param character: Character to pad with. Defaults to " ".

    """
    return align(val, "center", width=width, character=character)


def align_right(
    val: str,
    width: int | None = None,
    character: str = " ",
) -> str:
    """Right-align string to a given width.

    :param val: string to align
    :param width: Desired width. Defaults to width of the terminal.
    :param character: Character to pad with. Defaults to " ".

    """
    return align(val, "right", width=width, character=character)


def stylize(val: str, style: StyleType) -> str:
    """Apply an ANSI style to a string, preserving any existing styles.

    :param val: string to be styled
    :param style: style instance or style definition to apply.
    :return: the stylized string
    """
    # Convert to a Rich Text object to parse and preserve existing ANSI styles.
    text = Text.from_ansi(val)
    text.stylize(style)
    return ru.rich_text_to_string(text)


def strip_style(val: str) -> str:
    """Strip all ANSI style sequences (colors, bold, etc.) from a string.

    This targets SGR sequences specifically and leaves other terminal
    control codes intact.

    :param val: string to be stripped
    :return: the stripped string
    """
    return ru.ANSI_STYLE_SEQUENCE_RE.sub("", val)


def str_width(val: str) -> int:
    """Return the display width of a string.

    This is intended for single-line strings.
    Replace tabs with spaces before calling this.

    :param val: the string being measured
    :return: width of the string when printed to the terminal
    """
    text = Text.from_ansi(val)
    return text.cell_len


def is_quoted(val: str) -> bool:
    """Check if a string is quoted.

    :param val: the string being checked for quotes
    :return: True if a string is quoted
    """
    from . import constants

    return len(val) > 1 and val[0] == val[-1] and val[0] in constants.QUOTES


def quote(val: str) -> str:
    """Quote a string."""
    quote = "'" if '"' in val else '"'

    return quote + val + quote


def quote_if_needed(val: str) -> str:
    """Quote a string if it contains spaces and isn't already quoted."""
    if is_quoted(val) or " " not in val:
        return val

    return quote(val)


def strip_quotes(val: str) -> str:
    """Strip outer quotes from a string.

     Applies to both single and double quotes.

    :param val:  string to strip outer quotes from
    :return: same string with potentially outer quotes stripped
    """
    if is_quoted(val):
        val = val[1:-1]
    return val


def norm_fold(val: str) -> str:
    """Normalize and casefold Unicode strings for saner comparisons.

    :param val: input unicode string
    :return: a normalized and case-folded version of the input string
    """
    import unicodedata

    return unicodedata.normalize("NFC", val).casefold()


def common_prefix(m: Sequence[str]) -> str:
    """Return the longest common leading component of a list of strings.

    This is a replacement for os.path.commonprefix which is deprecated in Python 3.15.

    :param m: list of strings
    :return: common prefix
    """
    if not m:
        return ""

    s1 = min(m)
    s2 = max(m)
    for i, c in enumerate(s1):
        if i >= len(s2) or c != s2[i]:
            return s1[:i]
    return s1
